import { JSONValue } from "./types";
export type HttpMethod = 'DELETE' | 'GET' | 'PATCH' | 'POST' | 'PUT';
export type FetchDataTypesAllowed = JSONValue | object;
export type FetchSettings<Tdata extends FetchDataTypesAllowed | undefined = undefined> = {
    url: string;
    method: HttpMethod;
    data?: Tdata;
    fname?: string;
    bearerToken?: string;
    statusCodesToBypassErrorHandler?: number[];
};
/**
 * An HTTP header to support JSON API calls. An optional Bearer token can be provided as well.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns A JSON ready header for HTTP calls.
 */
export declare function getHttpHeaderJson(bearerToken?: string): Headers;
/**
 * Makes an HTTP call to an API using the given HTTP method.
 * @param url The URL endpoint of the API call.
 * @param method The HTTP method to be used.
 * @param data Optional body to send with the request. Can be a JSON object or a string.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned Response object in a Promise.
 */
export declare function fetchHttp<Tdata extends FetchDataTypesAllowed | undefined = undefined>(url: string, method: HttpMethod, data?: Tdata, fname?: string, bearerToken?: string, settings?: FetchSettings<Tdata>): Promise<Response>;
export declare function fetchData<Tdata extends FetchDataTypesAllowed>(url: string, method: HttpMethod, data?: Tdata, fname?: string, bearerToken?: string, settings?: FetchSettings<Tdata>): Promise<string>;
export declare function fetchJson<Tdata extends FetchDataTypesAllowed | undefined, Tret>(url: string, method: HttpMethod, data?: Tdata, fname?: string, bearerToken?: string, settings?: FetchSettings<Tdata>): Promise<Tret>;
/**
 * DELETEs data to an API using an HTTP DELETE.
 * @param url The URL endpoint of the API call.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @param fname The callers function name for outputting in potential error calls.
 * @returns The returned Response object in a Promise.
 */
export declare function fetchDelete(url: string, fname?: string, bearerToken?: string, settings?: FetchSettings): Promise<Response>;
/**
 * DELETEs data to an API using an HTTP DELETE.
 * @param url The URL endpoint of the API call.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @param fname The callers function name for outputting in potential error calls.
 * @returns The returned Response object in a Promise.
 */
export declare function fetchDeleteJson<Tdata extends FetchDataTypesAllowed, Tret = undefined>(url: string, fname?: string, bearerToken?: string, data?: Tdata, settings?: FetchSettings<Tdata>): Promise<Tret>;
/**
 * Fetches data from an API using an HTTP GET.
 * Returns the JSON data from the API call.
 * @param url The URL endpoint of the API call.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned JSON object.
 */
export declare function fetchGet<Tret>(url: string, fname?: string, bearerToken?: string, settings?: FetchSettings): Promise<Tret>;
/**
 * PATCHs data to an API using an HTTP POST.
 * @param url The URL endpoint of the API call.
 * @param data The object of the data to pass to the API.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned Response object in a Promise.
 */
export declare function fetchPatch<Tdata extends FetchDataTypesAllowed, Tret = undefined>(url: string, data: Tdata, fname?: string, bearerToken?: string, settings?: FetchSettings<Tdata>): Promise<Tret>;
/**
 * Fetches data from an API using an HTTP POST.
 * Returns the JSON data from the API call.
 * @param url The URL endpoint of the API call.
 * @param data The object of the data to pass to the API.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned JSON object.
 */
export declare function fetchPost<Tdata extends FetchDataTypesAllowed | undefined, Tret = undefined>(url: string, data?: Tdata, fname?: string, bearerToken?: string, settings?: FetchSettings<Tdata>): Promise<Tret>;
/**
 * PUTs data to an API using an HTTP POST.
 * @param url The URL endpoint of the API call.
 * @param data The object of the data to pass to the API.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned Response object in a Promise.
 */
export declare function fetchPut<Tdata extends FetchDataTypesAllowed, Tret = undefined>(url: string, data: Tdata, fname?: string, bearerToken?: string, settings?: FetchSettings<Tdata>): Promise<Tret>;
