"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.fetchPut = exports.fetchPost = exports.fetchPatch = exports.fetchGet = exports.fetchDeleteJson = exports.fetchDelete = exports.fetchJson = exports.fetchData = exports.fetchHttp = exports.getHttpHeaderJson = void 0;
const exception_types_1 = require("./exception-types");
const skky_1 = require("./skky");
/**
 * An HTTP header to support JSON API calls. An optional Bearer token can be provided as well.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns A JSON ready header for HTTP calls.
 */
function getHttpHeaderJson(bearerToken) {
    const headers = new Headers({ "Content-Type": "application/json" });
    if ((0, skky_1.hasData)(bearerToken)) {
        headers.append("Authorization", `Bearer ${bearerToken}`);
    }
    return headers;
}
exports.getHttpHeaderJson = getHttpHeaderJson;
/**
 * Makes an HTTP call to an API using the given HTTP method.
 * @param url The URL endpoint of the API call.
 * @param method The HTTP method to be used.
 * @param data Optional body to send with the request. Can be a JSON object or a string.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned Response object in a Promise.
 */
function fetchHttp(url, method, data, fname, bearerToken, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!fname || !(0, skky_1.hasData)(fname)) {
            fname = 'fetchHttp';
        }
        if (!(0, skky_1.hasData)(url)) {
            throw new Error(`${fname} passed an empty URL.`);
        }
        let response;
        try {
            const req = {
                method,
                headers: getHttpHeaderJson(bearerToken),
            };
            if (data && (0, skky_1.hasData)(data)) {
                req.body = (0, skky_1.isObject)(data) || (0, skky_1.isArray)(data) ? JSON.stringify(data) : String(data);
            }
            response = yield fetch(url, req);
        }
        catch (err) {
            if (err instanceof exception_types_1.GrayArrowExceptionHttp) {
                throw err;
            }
            throw new exception_types_1.GrayArrowExceptionHttp(err.message, fname);
        }
        if (!response) {
            throw new exception_types_1.GrayArrowExceptionHttp(`${fname}: NO response in HTTP ${method} to URL: ${url}.`, fname);
        }
        if (!response.ok) {
            let captureResponse;
            if (401 === response.status) {
                try {
                    captureResponse = yield response.json();
                }
                catch (jsonerr) { }
            }
            throw new exception_types_1.GrayArrowExceptionHttp(`${fname}: Error in HTTP ${method} to URL: ${url} with status code ${response.status}.`, fname, { response, captureResponse });
        }
        return response;
    });
}
exports.fetchHttp = fetchHttp;
function fetchData(url, method, data, fname, bearerToken, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        const resp = yield fetchHttp(url, method, data, fname, bearerToken, settings);
        return yield resp.text();
    });
}
exports.fetchData = fetchData;
function fetchJson(url, method, data, fname, bearerToken, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        const resp = yield fetchHttp(url, method, data, fname, bearerToken, settings);
        return yield resp.json();
    });
}
exports.fetchJson = fetchJson;
/**
 * DELETEs data to an API using an HTTP DELETE.
 * @param url The URL endpoint of the API call.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @param fname The callers function name for outputting in potential error calls.
 * @returns The returned Response object in a Promise.
 */
function fetchDelete(url, fname, bearerToken, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        return fetchHttp(url, "DELETE", undefined, fname, bearerToken, settings);
    });
}
exports.fetchDelete = fetchDelete;
/**
 * DELETEs data to an API using an HTTP DELETE.
 * @param url The URL endpoint of the API call.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @param fname The callers function name for outputting in potential error calls.
 * @returns The returned Response object in a Promise.
 */
function fetchDeleteJson(url, fname, bearerToken, data, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        return fetchJson(url, "DELETE", data, fname, bearerToken, settings);
    });
}
exports.fetchDeleteJson = fetchDeleteJson;
/**
 * Fetches data from an API using an HTTP GET.
 * Returns the JSON data from the API call.
 * @param url The URL endpoint of the API call.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned JSON object.
 */
function fetchGet(url, fname, bearerToken, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        return fetchJson(url, "GET", undefined, fname, bearerToken, settings);
    });
}
exports.fetchGet = fetchGet;
/**
 * PATCHs data to an API using an HTTP POST.
 * @param url The URL endpoint of the API call.
 * @param data The object of the data to pass to the API.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned Response object in a Promise.
 */
function fetchPatch(url, data, fname, bearerToken, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        return fetchJson(url, "PATCH", data, fname, bearerToken, settings);
    });
}
exports.fetchPatch = fetchPatch;
/**
 * Fetches data from an API using an HTTP POST.
 * Returns the JSON data from the API call.
 * @param url The URL endpoint of the API call.
 * @param data The object of the data to pass to the API.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned JSON object.
 */
function fetchPost(url, data, fname, bearerToken, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        return fetchJson(url, "POST", data, fname, bearerToken, settings);
    });
}
exports.fetchPost = fetchPost;
/**
 * PUTs data to an API using an HTTP POST.
 * @param url The URL endpoint of the API call.
 * @param data The object of the data to pass to the API.
 * @param fname The callers function name for outputting in potential error calls.
 * @param bearerToken An optional security token to add as Authorization to the HTTP header.
 * @returns The returned Response object in a Promise.
 */
function fetchPut(url, data, fname, bearerToken, settings) {
    return __awaiter(this, void 0, void 0, function* () {
        return fetchJson(url, "PUT", data, fname, bearerToken, settings);
    });
}
exports.fetchPut = fetchPut;
